from academia_mcp.tools.md_to_pdf import convert_md_to_latex, md_to_pdf


TEST_MD_BASIC = """
# Markdown syntax guide

## Headers

### This is a Heading h3
###### This is a Heading h6

## Emphasis

*This text will be italic*
_This will also be italic_

**This text will be bold**
__This will also be bold__

_You **can** combine them_
"""

TEST_MD_TABLES = """
## Tables

| Left columns  | Right columns |
| ------------- |:-------------:|
| left foo      | right foo     |
| left bar      | right bar     |
| left baz      | right baz     |
"""

TEST_MD_LINKS = """
## Links

You may be using [Markdown Live Preview](https://markdownlivepreview.com/).
"""

TEST_MD_IMAGES = """
## Images

![This is an alt text.](/image/sample.webp "This is a sample image.")
"""

TEST_MD_LISTS = """
## Lists

### Unordered

* Item 1
* Item 2
* Item 2a
* Item 2b
    * Item 3a
    * Item 3b

### Ordered

1. Item 1
2. Item 2
3. Item 3
    1. Item 3a
    2. Item 3b
"""


def test_md_to_latex_basic() -> None:
    latex = convert_md_to_latex(TEST_MD_BASIC)
    assert "\\documentclass{article}" in latex
    assert "\\title{Markdown syntax guide}" in latex
    assert "\\section{Headers}" in latex
    assert "\\section{Emphasis}" in latex
    assert "\\subsection{This is a Heading h3}" in latex
    assert "\\emph{This text will be italic}" in latex
    assert "\\emph{This will also be italic}" in latex
    assert "\\textbf{This text will be bold}" in latex
    assert "\\textbf{This will also be bold}" in latex
    assert "\\emph{You \\textbf{can} combine them}" in latex


def test_md_to_latex_tables() -> None:
    latex = convert_md_to_latex(TEST_MD_TABLES)
    assert "\\begin{table}[h]" in latex
    assert "\\begin{tabular}{|l|l|}" in latex
    assert "left foo & right foo" in latex


def test_md_to_latex_links() -> None:
    latex = convert_md_to_latex(TEST_MD_LINKS)
    assert (
        "You may be using \\href{https://markdownlivepreview.com/}{Markdown Live Preview}." in latex
    )


def test_md_to_latex_images() -> None:
    latex = convert_md_to_latex(TEST_MD_IMAGES)
    assert "\\begin{figure}" in latex
    assert "\\includegraphics[width=\\linewidth]{/image/sample.webp}" in latex
    assert "\\caption{This is an alt text.}" in latex
    assert "\\end{figure}" in latex


def test_md_to_latex_lists() -> None:
    latex = convert_md_to_latex(TEST_MD_LISTS)
    assert "\\begin{itemize}" in latex
    assert "\\end{itemize}" in latex
    assert "\\begin{enumerate}" in latex
    assert "\\end{enumerate}" in latex
    assert "\\item Item 3a" in latex


def test_md_to_pdf() -> None:
    result = md_to_pdf(TEST_MD_LISTS, "output")
    assert "Compilation successful" in result



import json

from academia_mcp.tools import s2_get_citations, s2_get_references


def test_s2_citations_pingpong() -> None:
    citations = json.loads(s2_get_citations("2409.06820"))
    assert citations["total_count"] >= 1
    assert "2502.18308" in str(citations["results"])


def test_s2_citations_transformers() -> None:
    try:
        citations = json.loads(s2_get_citations("1706.03762"))
        assert citations["total_count"] >= 100000
    except Exception:
        pass


def test_s2_citations_reversed() -> None:
    try:
        citations = json.loads(s2_get_references("1706.03762"))
        assert citations["total_count"] <= 100
    except Exception:
        pass


def test_s2_citations_versions() -> None:
    citations = json.loads(s2_get_citations("2409.06820v4"))
    assert citations["total_count"] >= 1



from academia_mcp.tools import document_qa, arxiv_download


DOCUMENT1 = """
The dominant sequence transduction models are based on complex recurrent or convolutional
neural networks in an encoder-decoder configuration. The best performing models also connect
the encoder and decoder through an attention mechanism. We propose a new simple network architecture,
the Transformer, based solely on attention mechanisms, dispensing with recurrence and convolutions
entirely. Experiments on two machine translation tasks show these models to be superior in quality
while being more parallelizable and requiring significantly less time to train. Our model achieves
28.4 BLEU on the WMT 2014 English-to-German translation task, improving over the existing best results,
including ensembles by over 2 BLEU. On the WMT 2014 English-to-French translation task, our model
establishes a new single-model state-of-the-art BLEU score of 41.8 after training for 3.5 days on
eight GPUs, a small fraction of the training costs of the best models from the literature.
We show that the Transformer generalizes well to other tasks by applying it successfully to
English constituency parsing both with large and limited training data.
"""


async def test_document_qa_base() -> None:
    answer = await document_qa(
        question="What is BLEU on the WMT 2014 English-to-German translation task?",
        document=DOCUMENT1,
    )
    assert "28.4" in answer


async def test_document_qa_real_question() -> None:
    questions = "What is the best model for the Russian language according to the role-play benchmark and its final score?"
    document = arxiv_download("2409.06820")
    answer = await document_qa(question=questions, document=document)
    assert "4.62" in answer or "4.68" in answer



import os
import json
from typing import Optional

from academia_mcp.utils import post_with_retries, get_with_retries


EXA_SEARCH_URL = "https://api.exa.ai/search"
TAVILY_SEARCH_URL = "https://api.tavily.com/search"
BRAVE_SEARCH_URL = "https://api.search.brave.com/res/v1/web/search"
EXCLUDE_DOMAINS = ["chatpaper.com"]


def web_search(
    query: str,
    limit: Optional[int] = 20,
    provider: Optional[str] = "tavily",
) -> str:
    """
    Search the web using Exa Search or Tavily and return normalized results.

    Returns a JSON object serialized to a string. The structure is: {"results": [...]}
    Every item in the "results" has at least the following fields: ("title", "url")
    Use `json.loads` to deserialize the result if you want to get specific fields.

    Args:
        query: The search query, required.
        limit: The maximum number of items to return. 20 by default, maximum 25.
        provider: The provider to use. "exa", "tavily" or "brave". "tavily" by default.
    """
    assert provider in (
        "exa",
        "tavily",
        "brave",
    ), "Error: provider must be either 'exa', 'tavily' or 'brave'"

    if provider == "exa":
        return exa_web_search(query, limit)
    if provider == "brave":
        return brave_web_search(query, limit)
    return tavily_web_search(query, limit)


def tavily_web_search(query: str, limit: Optional[int] = 20) -> str:
    """
    Search the web using Tavily and return results.

    Returns a JSON object serialized to a string. The structure is: {"results": [...]}
    Every item in the "results" has at least the following fields: ("title", "url")
    Use `json.loads` to deserialize the result if you want to get specific fields.

    Args:
        query: The search query, required.
        limit: The maximum number of items to return. 20 by default, maximum 25.
    """
    assert isinstance(query, str), "Error: Your search query must be a string"
    assert query.strip(), "Error: Your query should not be empty"
    assert isinstance(limit, int), "Error: limit should be an integer"
    assert 0 < limit <= 25, "Error: limit should be between 1 and 25"

    key = os.getenv("TAVILY_API_KEY", "")
    assert key, "Error: TAVILY_API_KEY is not set and no api_key was provided"
    payload = {
        "query": query,
        "max_results": limit,
        "auto_parameters": True,
        "exclude_domains": EXCLUDE_DOMAINS,
    }
    response = post_with_retries(TAVILY_SEARCH_URL, payload, key)
    results = response.json()["results"]
    for result in results:
        content = " ".join(result["content"].split(" ")[:40])
        content = content.strip("., ")
        result["content"] = content
        result.pop("raw_content", None)
        result.pop("score", None)
    return json.dumps({"results": results}, ensure_ascii=False)


def exa_web_search(query: str, limit: Optional[int] = 20) -> str:
    """
    Search the web using Exa and return results.

    Returns a JSON object serialized to a string. The structure is: {"results": [...]}
    Every item in the "results" has at least the following fields: ("title", "url")
    Use `json.loads` to deserialize the result if you want to get specific fields.

    Args:
        query: The search query, required.
        limit: The maximum number of items to return. 20 by default, maximum 25.
    """
    assert isinstance(query, str), "Error: Your search query must be a string"
    assert query.strip(), "Error: Your query should not be empty"
    assert isinstance(limit, int), "Error: limit should be an integer"
    assert 0 < limit <= 25, "Error: limit should be between 1 and 25"

    key = os.getenv("EXA_API_KEY", "")
    assert key, "Error: EXA_API_KEY is not set and no api_key was provided"
    payload = {
        "query": query,
        "type": "auto",
        "numResults": limit,
        "context": False,
        "excludeDomains": EXCLUDE_DOMAINS,
        "contents": {
            "text": False,
            "highlights": {
                "numSentences": 5,
            },
            "context": False,
        },
    }

    response = post_with_retries(EXA_SEARCH_URL, payload, key)
    results = response.json()["results"]
    return json.dumps({"results": results}, ensure_ascii=False)


def brave_web_search(query: str, limit: Optional[int] = 20) -> str:
    """
    Search the web using Brave and return results.

    Returns a JSON object serialized to a string. The structure is: {"results": [...]}
    Every item in the "results" has at least the following fields: ("title", "url")
    Use `json.loads` to deserialize the result if you want to get specific fields.

    Args:
        query: The search query, required.
        limit: The maximum number of items to return. 20 by default, maximum 20.
    """
    assert isinstance(query, str), "Error: Your search query must be a string"
    assert query.strip(), "Error: Your query should not be empty"
    assert isinstance(limit, int), "Error: limit should be an integer"
    assert 0 < limit <= 20, "Error: limit should be between 1 and 20"

    key = os.getenv("BRAVE_API_KEY", "")
    assert key, "Error: BRAVE_API_KEY is not set and no api_key was provided"
    payload = {
        "q": query,
        "count": limit,
    }
    response = get_with_retries(BRAVE_SEARCH_URL, key, params=payload)
    results = response.json()["web"]["results"]
    return json.dumps({"results": results}, ensure_ascii=False)



import re
import os
import json
from typing import Optional

from markdownify import markdownify  # type: ignore

from academia_mcp.utils import get_with_retries, post_with_retries

EXA_CONTENTS_URL = "https://api.exa.ai/contents"
AVAILABLE_PROVIDERS = ("basic", "exa")


def visit_webpage(url: str, provider: Optional[str] = "basic") -> str:
    """
    Visit a webpage and return the content.

    Returns a JSON object serialized to a string. The structure is: {"url": "...", "text": "..."}
    Use `json.loads` to deserialize the result if you want to get specific fields.
    Use "exa" provider in case "basic" fails.

    Args:
        url: The URL of the webpage to visit.
        provider: The provider to use. Available providers: "basic" (default) or "exa".
    """
    assert (
        provider in AVAILABLE_PROVIDERS
    ), f"Invalid provider: {provider}. Available providers: {AVAILABLE_PROVIDERS}"

    if provider == "exa":
        key = os.getenv("EXA_API_KEY", "")
        assert key, "Error: EXA_API_KEY is not set and no api_key was provided"
        payload = {
            "urls": [url],
            "text": True,
        }
        response = post_with_retries(EXA_CONTENTS_URL, payload=payload, api_key=key)
        return json.dumps(response.json()["results"][0])

    assert provider == "basic"
    response = get_with_retries(url)
    markdown_content = markdownify(response.text).strip()
    markdown_content = re.sub(r"\n{3,}", "\n\n", markdown_content)
    return json.dumps({"id": url, "text": markdown_content})



import os
from typing import List, Dict, Any

from pydantic import BaseModel
from openai import AsyncOpenAI
from openai.types.chat.chat_completion_message import ChatCompletionMessage


class ChatMessage(BaseModel):  # type: ignore
    role: str
    content: str | List[Dict[str, Any]]


ChatMessages = List[ChatMessage]


async def llm_acall(model_name: str, prompt: str) -> str:
    key = os.getenv("OPENROUTER_API_KEY", "")
    assert key, "Please set OPENROUTER_API_KEY in the environment variables"
    base_url = os.getenv("BASE_URL", "https://openrouter.ai/api/v1")

    messages: ChatMessages = [
        ChatMessage(role="user", content=prompt),
    ]
    client = AsyncOpenAI(base_url=base_url, api_key=key)
    response: ChatCompletionMessage = (
        (
            await client.chat.completions.create(
                model=model_name,
                messages=messages,
                temperature=0.0,
            )
        )
        .choices[0]
        .message
    )
    assert response.content, "Response content is None"
    return response.content



import os
import socket
from typing import Optional, Literal

import fire  # type: ignore
from mcp.server.fastmcp import FastMCP
from dotenv import load_dotenv

from .tools.arxiv_search import arxiv_search
from .tools.arxiv_download import arxiv_download
from .tools.s2_citations import s2_get_citations, s2_get_references
from .tools.hf_datasets_search import hf_datasets_search
from .tools.anthology_search import anthology_search
from .tools.document_qa import document_qa
from .tools.md_to_pdf import md_to_pdf
from .tools.web_search import web_search, tavily_web_search, exa_web_search, brave_web_search
from .tools.visit_webpage import visit_webpage
from .tools.bitflip import (
    extract_bitflip_info,
    generate_research_proposal,
    score_research_proposals,
)


load_dotenv()


def find_free_port() -> int:
    for port in range(5000, 6001):
        try:
            with socket.socket() as s:
                s.bind(("", port))
                return port
        except Exception:
            continue
    return 5000


def run(
    host: str = "0.0.0.0",
    port: Optional[int] = None,
    mount_path: str = "/",
    streamable_http_path: str = "/mcp",
    transport: Literal["stdio", "sse", "streamable-http"] = "streamable-http",
) -> None:
    server = FastMCP(
        "Academia MCP",
        stateless_http=True,
        streamable_http_path=streamable_http_path,
        mount_path=mount_path,
    )

    server.add_tool(arxiv_search)
    server.add_tool(arxiv_download)
    server.add_tool(s2_get_citations)
    server.add_tool(s2_get_references)
    server.add_tool(hf_datasets_search)
    server.add_tool(anthology_search)
    server.add_tool(md_to_pdf)
    server.add_tool(visit_webpage)
    server.add_tool(extract_bitflip_info)
    server.add_tool(generate_research_proposal)
    server.add_tool(score_research_proposals)

    if os.getenv("TAVILY_API_KEY"):
        server.add_tool(tavily_web_search)
    if os.getenv("EXA_API_KEY"):
        server.add_tool(exa_web_search)
    if os.getenv("BRAVE_API_KEY"):
        server.add_tool(brave_web_search)
    if os.getenv("EXA_API_KEY") or os.getenv("BRAVE_API_KEY") or os.getenv("TAVILY_API_KEY"):
        server.add_tool(web_search)
    if os.getenv("OPENROUTER_API_KEY"):
        server.add_tool(document_qa)

    if port is None:
        port = find_free_port()
    server.settings.port = port
    server.settings.host = host
    server.run(transport=transport)


if __name__ == "__main__":
    fire.Fire(run)



import pytest
from unittest import TestCase

from academia_mcp.utils import extract_json


class TestJSONExtractor(TestCase):
    def test_simple_json_code_block(self) -> None:
        text = """Here's the data:
        ```json
        {"name": "John", "age": 30}
        ```
        """
        expected = {"name": "John", "age": 30}
        result = extract_json(text)
        assert result == expected

    def test_json_code_block_case_insensitive(self) -> None:
        text = """```JSON
        {"status": "success"}
        ```"""
        expected = {"status": "success"}
        result = extract_json(text)
        assert result == expected

    def test_generic_code_block(self) -> None:
        text = """Here's the response:
        ```
        {"message": "Hello World", "code": 200}
        ```
        """
        expected = {"message": "Hello World", "code": 200}
        result = extract_json(text)
        assert result == expected

    def test_direct_json_parsing(self) -> None:
        text = '{"direct": true, "value": 42}'
        expected = {"direct": True, "value": 42}
        result = extract_json(text)
        assert result == expected

    def test_json_array(self) -> None:
        text = """The items are:
        ```json
        [1, 2, 3, {"name": "test"}]
        ```
        """
        expected = [1, 2, 3, {"name": "test"}]
        result = extract_json(text)
        assert result == expected

    def test_nested_json_objects(self) -> None:
        text = """```json
        {
            "user": {
                "id": 123,
                "profile": {
                    "name": "Alice",
                    "settings": {"theme": "dark"}
                }
            },
            "metadata": {"version": "1.0"}
        }
        ```"""
        expected = {
            "user": {"id": 123, "profile": {"name": "Alice", "settings": {"theme": "dark"}}},
            "metadata": {"version": "1.0"},
        }
        result = extract_json(text)
        assert result == expected

    def test_json_with_trailing_comma(self) -> None:
        text = """{
            "name": "test",
            "values": [1, 2, 3,],
            "nested": {"a": 1,},
        }"""
        expected = {"name": "test", "values": [1, 2, 3], "nested": {"a": 1}}
        result = extract_json(text)
        assert result == expected

    def test_json_with_single_quotes(self) -> None:
        text = """json: {
            'name': 'John',
            'age': 30,
            'active': true
        }"""
        expected = {"name": "John", "age": 30, "active": True}
        result = extract_json(text)
        assert result == expected

    def test_json_with_comments(self) -> None:
        text = """{
            "name": "test", // This is a comment
            /* Multi-line
               comment */
            "value": 42
        }"""
        expected = {"name": "test", "value": 42}
        result = extract_json(text)
        assert result == expected

    def test_json_with_prefix_text(self) -> None:
        prefixes = [
            "json: ",
            "JSON: ",
            "Here is the JSON: ",
            "The JSON is: ",
            "Result: ",
            "Output: ",
            "Response: ",
        ]

        base_json = {"test": "value"}

        for prefix in prefixes:
            text = prefix + '{"test": "value"}'
            result = extract_json(text)
            assert result == base_json, f"Failed for prefix: {prefix}"

    def test_empty_input(self) -> None:
        with pytest.raises(AssertionError):
            extract_json("")

        with pytest.raises(AssertionError):
            extract_json(None)  # type: ignore

    def test_non_string_input(self) -> None:
        with pytest.raises(AssertionError):
            extract_json(123)  # type: ignore

        with pytest.raises(AssertionError):
            extract_json(["not", "a", "string"])  # type: ignore

    def test_no_json_found(self) -> None:
        text = "This is just plain text with no JSON whatsoever."
        result = extract_json(text)
        assert result is None

    def test_malformed_json_in_code_block(self) -> None:
        text = """```json
        {"name": "test", "invalid": }
        ```"""

        result = extract_json(text)
        assert result is None

    def test_json_with_special_characters(self) -> None:
        text = """```json
        {
            "message": "Hello 🌍",
            "path": "/users/test",
            "emoji": "😀",
            "unicode": "café"
        }
        ```"""

        expected = {"message": "Hello 🌍", "path": "/users/test", "emoji": "😀", "unicode": "café"}

        result = extract_json(text)
        assert result == expected

    def test_json_with_numbers_and_booleans(self) -> None:
        text = """```json
        {
            "integer": 42,
            "float": 3.14159,
            "boolean_true": true,
            "boolean_false": false,
            "null_value": null,
            "negative": -100
        }
        ```"""

        expected = {
            "integer": 42,
            "float": 3.14159,
            "boolean_true": True,
            "boolean_false": False,
            "null_value": None,
            "negative": -100,
        }

        result = extract_json(text)
        assert result == expected

    def test_complex_mixed_content(self) -> None:
        text = """
        Based on your request, here's the user data analysis:

        The system found 3 users in the database. Here's the detailed breakdown:

        ```json
        {
            "total_users": 3,
            "users": [
                {
                    "id": 1,
                    "name": "Alice Johnson",
                    "email": "alice@example.com",
                    "active": true,
                    "roles": ["user", "moderator"]
                },
                {
                    "id": 2,
                    "name": "Bob Smith",
                    "email": "bob@example.com",
                    "active": false,
                    "roles": ["user"]
                }
            ],
            "metadata": {
                "generated_at": "2024-01-15T10:30:00Z",
                "query_time_ms": 45
            }
        }
        ```

        This data was generated from the user management system.
        """

        result = extract_json(text)

        assert isinstance(result, dict)
        assert result["total_users"] == 3
        assert len(result["users"]) == 2
        assert "metadata" in result

        alice = result["users"][0]
        assert alice["name"] == "Alice Johnson"
        assert alice["roles"] == ["user", "moderator"]



import json

from academia_mcp.tools import web_search


def test_web_search_base() -> None:
    result = web_search("autoregressive models path-star graphs", limit=20)
    assert "The Mystery of the Pathological" in result
    results = json.loads(result)
    assert results
    assert "score" not in str(results)


def test_web_search_exa() -> None:
    result = web_search("autoregressive models path-star graphs", provider="exa", limit=10)
    assert result
    results = json.loads(result)
    assert results


def test_web_search_brave() -> None:
    result = web_search("autoregressive models path-star graphs", provider="brave", limit=10)
    assert "The Mystery of the Pathological" in result
    results = json.loads(result)
    assert results



import json

from academia_mcp.tools.bitflip import (
    extract_bitflip_info,
    generate_research_proposal,
    score_research_proposals,
)


async def test_bitflip_extract_info() -> None:
    arxiv_id = "2409.06820"
    result = json.loads(await extract_bitflip_info(arxiv_id))
    assert result is not None
    assert result["bit"]


async def test_bitflip_generate_research_proposal() -> None:
    arxiv_id = "2503.07826"
    bit = json.loads(await extract_bitflip_info(arxiv_id))["bit"]
    result = json.loads(await generate_research_proposal(bit=bit))
    assert result is not None
    assert result["flip"]


async def test_bitflip_score_research_proposals() -> None:
    arxiv_id = "2503.07826"
    bit = json.loads(await extract_bitflip_info(arxiv_id))["bit"]
    proposal1 = await generate_research_proposal(bit=bit)
    proposal2 = await generate_research_proposal(bit=bit)
    scores = json.loads(await score_research_proposals([proposal1, proposal2]))
    assert scores
    assert len(scores) == 2
    assert scores[0]["spark"] is not None
    assert scores[1]["spark"] is not None
    assert scores[0]["strengths"] is not None
    assert scores[1]["strengths"] is not None
    assert scores[0]["weaknesses"] is not None
    assert scores[1]["weaknesses"] is not None



# Based on
# https://github.com/SamuelSchmidgall/AgentLaboratory/blob/main/tools.py
# https://github.com/bytedance/pasa/blob/main/utils.py

import re
import json
import tempfile
from pathlib import Path
from typing import Any, List, Optional, Dict
from dataclasses import dataclass, field

import requests
import bs4
from markdownify import MarkdownConverter  # type: ignore
from pypdf import PdfReader

from academia_mcp.utils import get_with_retries


def download_pdf(url: str, output_path: Path) -> None:
    response = get_with_retries(url)
    response.raise_for_status()
    content_type = response.headers.get("content-type")
    assert content_type
    assert "application/pdf" in content_type.lower()
    with open(output_path.resolve(), "wb") as fp:
        fp.write(response.content)


def parse_pdf_file(pdf_path: Path) -> List[str]:
    # Why not Marker? Because it is too heavy.
    reader = PdfReader(str(pdf_path.resolve()))

    pages = []
    for page_number, page in enumerate(reader.pages, start=1):
        try:
            text = page.extract_text()
            prefix = f"## Page {page_number}\n\n"
            pages.append(prefix + text)
        except Exception:
            continue
    return pages


HTML_URL = "https://arxiv.org/html/{paper_id}"
ABS_URL = "https://arxiv.org/abs/{paper_id}"
PDF_URL = "https://arxiv.org/pdf/{paper_id}"
SECTION_STOP_WORDS = (
    "references",
    "acknowledgments",
    "about this document",
    "appendix",
)


@dataclass
class TOCEntry:
    level: int
    title: str
    html_id: Optional[str] = None
    subsections: List["TOCEntry"] = field(default_factory=list)

    def linearize(self) -> List["TOCEntry"]:
        entries = [self]
        for subsection in self.subsections:
            entries.extend(subsection.linearize())
        return entries

    def is_excluded(self) -> bool:
        return any(ss in self.title.lower() for ss in SECTION_STOP_WORDS)

    def to_str(self) -> str:
        final_items = []
        output_index = 0
        for entry in self.linearize():
            if entry.level <= 1:
                continue
            prefix = "  " * (entry.level - 2)
            suffix = ""
            if entry.level == 2 and not entry.is_excluded():
                suffix = f" (index in 'sections': {output_index})"
                output_index += 1
            final_items.append(prefix + entry.title + suffix)
        return "\n".join(final_items)


class ArxivHTMLConverter(MarkdownConverter):  # type: ignore
    def __init__(self, base_url: str, *args: Any, **kwargs: Any) -> None:
        self.base_url = base_url

        super().__init__(*args, **kwargs)

    def convert_cite(self, el: bs4.element.Tag, text: str, convert_as_inline: bool = False) -> str:
        citations = text.split(";")
        citations = [c.strip() for c in citations if c.strip()]
        fixed_citations = []
        for citation in citations:
            fixed_citation = citation
            parts = citation.split()
            year = parts[-1]
            if len(year) > 4 and year[0] == "(" and year[-1] == ")":
                fixed_citation = f'{" ".join(parts[:-1])}, {year[1:-1]}'
            fixed_citations.append(fixed_citation)
        return f'({"; ".join(fixed_citations)})'

    def convert_sup(self, *args: Any, **kwargs: Any) -> str:
        return ""

    def convert_span(self, el: bs4.element.Tag, text: str, convert_as_inline: bool = False) -> str:
        if "class" not in el:
            return text
        if "ltx_tag_item" in el["class"]:
            return ""
        if "ltx_note_outer" in el["class"]:
            return f" (Footnote {text})"
        if "ltx_tag_note" in el["class"]:
            return text + ": "
        return text

    def convert_figure(
        self, el: bs4.element.Tag, text: str, convert_as_inline: bool = False
    ) -> str:
        if el.img:
            link = el.img.get("src")
            caption = "Figure"
            if el.figcaption:
                caption = el.figcaption.text
            return f"\n\n![{caption}]({self.base_url}/{link})\n\n"
        elif el.table:
            caption = "Table"
            if el.figcaption:
                caption = el.figcaption.text
            table_text = self.process_tag(el.table, convert_as_inline=False)
            return f"\n\n{caption}\n\n{table_text}\n\n"
        return text


def _generate_toc(soup: bs4.element.Tag) -> TOCEntry:
    stack = [TOCEntry(level=0, title="ROOT", html_id=None)]
    heading_tags = {"h1": 1, "h2": 2, "h3": 3, "h4": 4, "h5": 5}
    for tag in soup.find_all(heading_tags.keys()):
        level = heading_tags[tag.name]
        while stack[-1].level >= level:
            stack.pop()
        parent_entry = stack[-1]
        section = tag.find_parent("section", id=True)
        if not section:
            continue
        section_id = section.get("id")
        assert section_id is not None
        title = tag.get_text().strip()
        new_entry = TOCEntry(level=level, title=title, html_id=section_id)
        parent_entry.subsections.append(new_entry)
        stack.append(new_entry)
    return stack[0]


def _convert_soup_to_md(soup: bs4.element.Tag, url: str) -> str:
    converter = ArxivHTMLConverter(base_url=url, strip=["div", "a"], heading_style="ATX")
    md_content: str = converter.convert_soup(soup)
    md_content = md_content.replace("\xa0", " ")
    md_content = "\n".join([line.strip() for line in md_content.split("\n")])
    lines = [line.strip() for line in md_content.split("\n\n") if line.strip()]
    md_content = "\n\n".join(lines)
    return md_content


def _build_by_toc(toc: TOCEntry, soup: bs4.element.Tag, url: str) -> List[str]:
    final_sections = []
    for toc_entry in toc.linearize():
        if toc_entry.level == 2 and not toc_entry.is_excluded():
            section = soup.find(id=toc_entry.html_id)
            assert isinstance(section, bs4.element.Tag)
            text = _convert_soup_to_md(section, url)
            final_sections.append(text)
    return final_sections


def _format_authors(authors: str) -> str:
    if not authors:
        return ""
    names = authors.split(",")
    names = [n.strip() for n in names if n.strip()]
    result = ", ".join(names[:3])
    if len(names) > 3:
        result += f", and {len(names) - 3} more authors"
    return result


def _parse_citation_metadata(metas: List[str]) -> Dict[str, Any]:
    metas = [item.replace("\n", " ") for item in metas]
    meta_string = " ".join(metas)
    authors, title, journal, year = "", "", "", None
    if len(metas) == 3:
        authors, title, journal = metas
    else:
        meta_string = re.sub(r"\.\s\d{4}[a-z]?\.", ".", meta_string)
        match = re.match(r"^(.*?\.\s)(.*?)(\.\s.*|$)", meta_string, re.DOTALL)
        if match:
            authors = match.group(1).strip() if match.group(1) else ""
            title = match.group(2).strip() if match.group(2) else ""
            journal = match.group(3).strip() if match.group(3) else ""
            if journal.startswith(". "):
                journal = journal[2:]

    if authors:
        parts = authors.strip(".").split(".")
        if parts and parts[-1].strip().isdigit():
            year = int(parts[-1].strip())
            authors = ".".join(parts[:-1])
        authors = _format_authors(authors)

    result = {
        "authors": authors,
        "year": year,
        "title": title,
        "journal": journal,
    }
    if not authors or not title:
        result["meta"] = meta_string
    return result


def _extract_citations(soup_biblist: bs4.element.Tag) -> List[Dict[str, Any]]:
    extracted = []
    for li in soup_biblist.find_all("li", recursive=False):
        metas = [x.text.strip() for x in li.find_all("span", class_="ltx_bibblock")]
        extracted.append(_parse_citation_metadata(metas))
    return extracted


def _parse_html(paper_id: str) -> Dict[str, Any]:
    url = HTML_URL.format(paper_id=paper_id)
    response = get_with_retries(url)
    response.raise_for_status()
    content = response.text

    soup = bs4.BeautifulSoup(content, features="lxml")
    article = soup.article
    assert article and isinstance(article, bs4.element.Tag)

    citations = []
    biblist_tag = article.find(class_="ltx_biblist")
    if biblist_tag and isinstance(biblist_tag, bs4.element.Tag):
        citations = _extract_citations(biblist_tag)

    toc = _generate_toc(article)
    sections = _build_by_toc(toc, article, url)
    return {
        "toc": toc.to_str(),
        "sections": sections,
        "citations": citations,
        "original_format": "html",
    }


def _parse_abs(paper_id: str) -> Dict[str, str]:
    url = ABS_URL.format(paper_id=paper_id)
    response = get_with_retries(url)
    response.raise_for_status()
    content = response.text

    soup = bs4.BeautifulSoup(content, features="lxml")
    title_tag = soup.find(class_="title")
    assert title_tag and isinstance(title_tag, bs4.element.Tag)
    title = title_tag.get_text().strip()
    title = title.replace("Title:", "")
    abstract_tag = soup.find(class_="abstract")
    assert abstract_tag and isinstance(abstract_tag, bs4.element.Tag)
    abstract = abstract_tag.get_text().strip()
    abstract = abstract.replace("Abstract:", "")
    return {"title": title, "abstract": abstract}


def _parse_pdf(paper_id: str) -> Dict[str, Any]:
    url = PDF_URL.format(paper_id=paper_id)
    with tempfile.NamedTemporaryFile(mode="w", delete=True) as temp_file:
        download_pdf(url, Path(temp_file.name))
        pages: List[str] = parse_pdf_file(Path(temp_file.name))
    return {
        "toc": "\n".join([f"Page {page_number}" for page_number in range(1, len(pages) + 1)]),
        "sections": pages,
        "citations": [],
        "original_format": "pdf",
    }


def arxiv_download(
    paper_id: str,
    include_citations: Optional[bool] = False,
    mode: Optional[str] = "html",
) -> str:
    """
    Downloads a paper from Arxiv and converts it to text.
    Use mode = "html" by default.
    Fall back to mode = "pdf" if there are any problems with the HTML version.

    Returns a JSON with a following structure:
    {
        "title": "...",
        "abstract": "...",
        "toc": "...",
        "sections": ["...", ...],
        "citations": [...]
    }
    Use `json.loads` to deserialize the result if you want to get specific fields.
    For example, `abstract = json.loads(arxiv_download("2409.06820v1"))`
    The "toc" key contains Table of Contents, that sometimes has indexing for sections.

    Args:
        paper_id: ID of the paper on Arxiv. For instance: 2409.06820v1
        include_citations: include "citations" in the result or not. False by default.
        mode: Which version of paper to use. Options: ["html", "pdf"]. "html" by default.
    """

    abs_meta = _parse_abs(paper_id)
    if mode == "html":
        try:
            content = _parse_html(paper_id)
        except requests.exceptions.HTTPError:
            content = _parse_pdf(paper_id)
    else:
        content = _parse_pdf(paper_id)

    if not include_citations and "citations" in content:
        content.pop("citations")

    return json.dumps({**abs_meta, **content}, ensure_ascii=False)



from .arxiv_search import arxiv_search
from .anthology_search import anthology_search
from .arxiv_download import arxiv_download
from .hf_datasets_search import hf_datasets_search
from .s2_citations import s2_get_references, s2_get_citations
from .document_qa import document_qa
from .md_to_pdf import md_to_pdf
from .web_search import web_search, tavily_web_search, exa_web_search, brave_web_search
from .visit_webpage import visit_webpage
from .bitflip import extract_bitflip_info, generate_research_proposal, score_research_proposals


__all__ = [
    "arxiv_search",
    "arxiv_download",
    "anthology_search",
    "s2_get_references",
    "s2_get_citations",
    "hf_datasets_search",
    "document_qa",
    "md_to_pdf",
    "web_search",
    "tavily_web_search",
    "exa_web_search",
    "brave_web_search",
    "visit_webpage",
    "extract_bitflip_info",
    "generate_research_proposal",
    "score_research_proposals",
]



import json
import re
from datetime import datetime
from typing import Optional, List, Dict, Any

from acl_anthology import Anthology


class AnthologySingleton:
    instance: Optional[Anthology] = None

    @classmethod
    def get(cls) -> Anthology:
        if cls.instance is None:
            cls.instance = Anthology.from_repo()
            cls.instance.load_all()
        return cls.instance


SORT_BY_OPTIONS = ("relevance", "submittedDate", "published")
SORT_BY_MAPPING = {"submittedDate": "published"}
SORT_ORDER_OPTIONS = ("ascending", "descending")


def _format_text_field(text: str) -> str:
    return " ".join([line.strip() for line in text.split() if line.strip()])


def _format_authors(authors: List[Any]) -> str:
    names = [f"{author.name.first} {author.name.last}" for author in authors]
    result = ", ".join(names[:3])
    if len(names) > 3:
        result += f", and {len(names) - 3} more authors"
    return result


def _format_date(date_str: str) -> str:
    try:
        return datetime.strptime(date_str, "%Y").strftime("%B %d, %Y")
    except ValueError:
        return date_str


def _clean_entry(entry: Any) -> Dict[str, Any]:
    return {
        "id": entry.full_id,
        "title": _format_text_field(entry.title.as_text()),
        "authors": _format_authors(entry.authors),
        "abstract": (_format_text_field(entry.abstract.as_text()) if entry.abstract else ""),
        "published": _format_date(entry.year),
        "categories": ", ".join(entry.venue_ids),
        "comment": entry.note if entry.note else "",
        "url": entry.pdf.url if entry.pdf else "",
    }


def _convert_to_year(date_str: str) -> int:
    try:
        return int(date_str[:4])
    except ValueError as e:
        raise ValueError("Invalid date format. Please use YYYY-MM-DD format.") from e


def _has_cyrillic(text: str) -> bool:
    return bool(re.search("[а-яА-Я]", text))


def _parse_query(query: str, paper: Any) -> bool:
    conditions = re.split(r"\s+(AND|OR|ANDNOT)\s+", query)
    result = False
    for i in range(0, len(conditions), 2):
        condition = conditions[i]
        field, value = condition.split(":", 1) if ":" in condition else ("ti", condition)
        value = value.lower().replace('"', "").replace("'", "")
        match field:
            case "ti":
                match_found = value in paper.title.as_text().lower()
            case "au":
                match_found = any(value in str(author).lower() for author in paper.authors)
            case "abs":
                match_found = paper.abstract and value in paper.abstract.as_text().lower()
            case "cat":
                match_found = any(value in cat.lower() for cat in paper.venue_ids)
            case "id":
                match_found = value in paper.full_id.lower()
            case _:
                match_found = False
        if i == 0:
            result = match_found
        else:
            operator = conditions[i - 1]
            if operator == "AND":
                result = result and match_found
            elif operator == "OR":
                result = result or match_found
            elif operator == "ANDNOT":
                result = result and not match_found
    return result


def anthology_search(
    query: str,
    offset: Optional[int] = 0,
    limit: Optional[int] = 5,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    sort_by: Optional[str] = "relevance",
    sort_order: Optional[str] = "descending",
    include_abstracts: Optional[bool] = False,
) -> str:
    """
    Search ACL Anthology papers with field-specific queries.

    Fields:
        all: (all fields), ti: (title), au: (author),
        abs: (abstract), cat: (category), id: (ID without version)

    Operators:
        AND, OR, ANDNOT

    Please always specify the fields. Search should always be field-specific.
    You can search for an exact match of an entire phrase by enclosing the phrase in double quotes.
    If you do not need an exact match of a phrase, use single terms with OR/AND.
    Boolean operators are strict. In most cases, you need OR and not AND.
    You can scroll all search results with the "offset" parameter.
    Do not include date constraints in the query: use "start_date" and "end_date" parameters instead.
    The names of authors should be in Latin script. For example, search "Ilya Gusev" instead of "Илья Гусев".

    Example queries:
        abs:"machine learning"
        au:"del maestro"
        au:vaswani AND ti:"attention is all"
        all:role OR all:playing OR all:"language model"
        (au:vaswani OR au:"del maestro") ANDNOT ti:attention

    Return a JSON object serialized to a string. The structure is:
    {"total_count": ..., "returned_count": ..., "offset": ..., "results": [...]}
    Every item in the "results" has the following fields:
    ("index", "id", "title", "authors", "abstract", "published", "updated", "categories", "comment")
    You can use `json.loads` to deserialize the result and get specific fields.

    Args:
        query: The search query, required.
        offset: The offset in search results. If it is 10, the first 10 items will be skipped. 0 by default.
        limit: The maximum number of items that will be returned. limit=5 by default, limit=10 is the maximum.
        start_date: Start date in %Y-%m-%d format. None by default.
        end_date: End date in %Y-%m-%d format. None by default.
        sort_by: 3 options to sort by: relevance, lastUpdatedDate, submittedDate. relevance by default.
        sort_order: 2 sort orders: ascending, descending. descending by default.
        include_abstracts: include abstracts in the result or not. False by default.
    """
    assert isinstance(query, str), "Error: Your search query must be a string"
    assert isinstance(offset, int), "Error: offset should be an integer"
    assert isinstance(limit, int), "Error: limit should be an integer"
    assert isinstance(sort_by, str), "Error: sort_by should be a string"
    assert isinstance(sort_order, str), "Error: sort_order should be a string"
    assert query.strip(), "Error: Your query should not be empty"
    assert sort_by in SORT_BY_OPTIONS, f"Error: sort_by should be one of {SORT_BY_OPTIONS}"
    assert (
        sort_order in SORT_ORDER_OPTIONS
    ), f"Error: sort_order should be one of {SORT_ORDER_OPTIONS}"
    assert offset >= 0, "Error: offset must be 0 or positive number"
    assert limit < 100, "Error: limit is too large, it should be less than 100"
    assert limit > 0, "Error: limit should be greater than 0"
    assert not _has_cyrillic(query), "Error: use only Latin script for queries"
    assert include_abstracts is not None, "Error: include_abstracts must be bool"

    sort_by = SORT_BY_MAPPING.get(sort_by, sort_by)

    singleton = AnthologySingleton.get()
    all_papers = [
        paper for paper in singleton.papers() if paper.abstract and str(paper.abstract).strip()
    ]

    if start_date or end_date:
        start_year = _convert_to_year(start_date) if start_date else 1900
        end_year = _convert_to_year(end_date) if end_date else datetime.now().year
        all_papers = [paper for paper in all_papers if start_year <= int(paper.year) <= end_year]

    filtered_papers = [paper for paper in all_papers if _parse_query(query, paper)]

    if sort_by == "published":
        filtered_papers.sort(key=lambda x: int(x.year), reverse=(sort_order == "descending"))

    paged_papers = filtered_papers[offset : offset + limit]
    clean_entries = [_clean_entry(entry) for entry in paged_papers]

    return json.dumps(
        {
            "total_count": len(filtered_papers),
            "returned_count": len(paged_papers),
            "offset": offset,
            "results": clean_entries,
        },
        ensure_ascii=False,
    )



from academia_mcp.tools import visit_webpage


def test_visit_webpage_basic() -> None:
    content = visit_webpage("https://example.com/")
    assert content is not None
    assert "Example Domain" in content
    assert "illustrative" in content


def test_visit_webpage_exa() -> None:
    content = visit_webpage("https://example.com/", provider="exa")
    assert content is not None
    assert "Example Domain" in content
    assert "illustrative" in content



import markdown
import re
import os
import subprocess
import shutil
import xml.dom.minidom
import xml.etree.ElementTree as etree
from typing import Optional, Any

from markdown.core import Markdown

from academia_mcp.files import get_workspace_dir


START_SINGLE_QUOTE_RE = re.compile(r"(^|\s|\")'")
START_DOUBLE_QUOTE_RE = re.compile(r"(^|\s|'|`)\"")
END_DOUBLE_QUOTE_RE = re.compile(r'"(,|\.|\s|$)')


MAIN_TEMPLATE = """\\documentclass{{article}}
\\usepackage[utf8]{{inputenc}}
\\usepackage[T1]{{fontenc}}
\\usepackage{{textcomp}}
\\usepackage{{amsmath}}
\\usepackage{{float}}
\\usepackage{{graphicx}}
\\usepackage{{enumitem}}
\\usepackage{{quoting}}
\\usepackage{{booktabs}}
\\usepackage{{caption}}
\\usepackage{{siunitx}}
\\sisetup{{
  group-separator = {{,}},
  output-decimal-marker = {{.}}
}}
\\usepackage{{hyperref}}

\\author{{Holosophos}}

\\begin{{document}}

{latex_content}

\\end{{document}}"""


IMAGE_TEMPLATE = """\\begin{{figure}}[H]
\\centering
\\includegraphics[width=\\linewidth]{{{src}}}
\\caption{{{alt}}}
\\end{{figure}}"""


TABLE_TEMPLATE = """
\\begin{{table}}[h]
\\begin{{tabular}}{{{descriptor}}}
{core}
\\hline
\\end{{tabular}}
\\\\[5pt]
\\caption{{{caption}}}
\\end{{table}}
"""

ITEMIZE_TEMPLATE = """
\\begin{{itemize}}
{content}
\\end{{itemize}}
"""

QUOTE_TEMPLATE = """
\\begin{{quotation}}
{content}
\\end{{quotation}}
"""

VERBATIM_TEMPLATE = """
\\begin{{verbatim}}
{content}
\\end{{verbatim}}
"""


MAKETITLE = """
% ----------------------------------------------------------------
\\maketitle
% ----------------------------------------------------------------
"""


def inline_html_latex(text: str) -> str:
    out = text
    if re.search(r"&ldquo;.*?&rdquo;", text, flags=re.DOTALL):
        out = out.replace("&ldquo;", "\\enquote{").replace("&rdquo;", "}")
    if re.search(r"&lsquo;.*?&rsquo;", text, flags=re.DOTALL):
        out = out.replace("&lsquo;", "\\enquote{").replace("&rsquo;", "}")
    if re.search(r"&ldquo;.*?&ldquo;", text, flags=re.DOTALL):
        out = out.replace("&ldquo;", "\\enquote{", 1).replace("&ldquo;", "}", 1)
    if re.search(r"&laquo;.*?&raquo;", text, flags=re.DOTALL):
        out = out.replace("&laquo;", "\\enquote{").replace("&raquo;", "}")
    out = out.replace("...", "\\dots")
    out = out.replace("&hellip;", "\\dots")
    out = out.replace("&ndash;", "--")
    out = out.replace("&mdash;", "---")
    out = out.replace("\\|", "|")
    return out


def unescape_html_entities(text: str) -> str:
    mapping = {
        "&amp;": "&",
        "&lt;": "<",
        "&gt;": ">",
        "&quot;": '"',
    }
    for k, v in mapping.items():
        text = text.replace(k, v)
    return text


def escape_latex_entities(text: str) -> str:
    out = unescape_html_entities(text)
    out = out.replace("%", "\\%")
    out = out.replace("&", "\\&")
    out = out.replace("#", "\\#")
    out = START_SINGLE_QUOTE_RE.sub(r"\g<1>`", out)
    out = START_DOUBLE_QUOTE_RE.sub(r"\g<1>``", out)
    out = END_DOUBLE_QUOTE_RE.sub(r"''\g<1>", out)
    return out


class LaTeXExtension(markdown.Extension):
    def __init__(self, configs: Optional[Any] = None) -> None:
        self.reset()

    def extendMarkdown(self, md: Markdown) -> None:
        self.md = md
        latex_tp = LaTeXTreeProcessor()
        math_pp = MathTextPostProcessor()
        table_pp = TableTextPostProcessor()
        image_pp = ImageTextPostProcessor()
        link_pp = LinkTextPostProcessor()
        unescape_html_pp = UnescapeHtmlTextPostProcessor()

        md.treeprocessors.register(latex_tp, "latex", 20)
        md.postprocessors.register(unescape_html_pp, "unescape_html", 20)
        md.postprocessors.register(math_pp, "math", 20)
        md.postprocessors.register(image_pp, "image", 20)
        md.postprocessors.register(table_pp, "table", 20)
        md.postprocessors.register(link_pp, "link", 20)

    def reset(self) -> None:
        pass


class LaTeXTreeProcessor(markdown.treeprocessors.Treeprocessor):
    def run(self, doc: etree.Element) -> None:
        latex_text = self.tolatex(doc)
        doc.clear()
        latex_node = etree.Element("plaintext")
        latex_node.text = latex_text
        doc.append(latex_node)

    def tolatex(self, ournode: etree.Element) -> str:
        buf = ""
        subcontent = ""

        if ournode.text:
            subcontent += escape_latex_entities(ournode.text)

        for child in list(ournode):
            subcontent += self.tolatex(child)

        tag = ournode.tag
        if tag == "h1":
            buf += "\n\\title{%s}\n" % subcontent
            buf += MAKETITLE
        elif tag == "h2":
            buf += "\n\n\\section{%s}\n" % subcontent
        elif tag == "h3":
            buf += "\n\n\\subsection{%s}\n" % subcontent
        elif tag == "h4":
            buf += "\n\\subsubsection{%s}\n" % subcontent
        elif tag == "hr":
            buf += "\\noindent\\makebox[\\linewidth]{\\rule{\\linewidth}{0.4pt}}"
        elif tag == "ul":
            buf += ITEMIZE_TEMPLATE.format(content=subcontent.strip())
        elif tag == "ol":
            buf += " \\begin{enumerate}"
            if "start" in ournode.attrib:
                start = int(ournode.attrib["start"]) - 1
                buf += "\\setcounter{enumi}{" + str(start) + "}"
            buf += f"\n{subcontent}\n\\end{{enumerate}}"
        elif tag == "li":
            buf += "\n  \\item %s" % subcontent.strip()
        elif tag == "blockquote":
            buf += QUOTE_TEMPLATE.format(content=subcontent.strip())
        elif tag == "pre":
            buf += VERBATIM_TEMPLATE.format(content=subcontent.strip())
        elif tag == "q":
            buf += "`%s'" % subcontent.strip()
        elif tag == "p":
            buf += "\n%s\n" % subcontent.strip()
        elif tag == "sup":
            buf += "\\footnote{%s}" % subcontent.strip()
        elif tag == "strong":
            buf += "\\textbf{%s}" % subcontent.strip()
        elif tag == "em":
            buf += "\\emph{%s}" % subcontent.strip()
        elif tag == "table":
            buf += "\n\n<table>%s</table>\n\n" % subcontent
        elif tag == "thead":
            buf += "<thead>%s</thead>" % subcontent
        elif tag == "tbody":
            buf += "<tbody>%s</tbody>" % subcontent
        elif tag == "tr":
            buf += "<tr>%s</tr>" % subcontent
        elif tag == "th":
            buf += "<th>%s</th>" % subcontent
        elif tag == "td":
            buf += "<td>%s</td>" % subcontent
        elif tag == "img":
            buf += '<img src="%s" alt="%s" />' % (
                ournode.get("src"),
                ournode.get("alt"),
            )
        elif tag == "a":
            href = ournode.get("href")
            assert href
            buf += '<a href="%s">%s</a>' % (
                escape_latex_entities(href),
                subcontent,
            )
        else:
            buf = subcontent

        if ournode.tail:
            buf += escape_latex_entities(ournode.tail)

        return buf


class Table2Latex:
    def convert_markdown_table(self, instr: str) -> str:
        lines = instr.split("\n")
        headers = lines[0].strip("|").split("|")
        cols = len(headers)
        buf = (
            "\\begin{table}[h]\n\\centering\n\\begin{tabular}{|"
            + "|".join(["l"] * cols)
            + "|}\n\\hline\n"
        )
        buf += (
            " & ".join([f"\\textbf{{{header.strip()}}}" for header in headers]) + " \\\\\n\\hline\n"
        )
        for line in lines[2:]:
            cells = line.strip("|").split("|")
            buf += " & ".join([cell.strip() for cell in cells]) + " \\\\\n\\hline\n"
        buf += "\\end{tabular}\n\\end{table}"
        return buf


class Img2Latex:
    def convert(self, instr: str) -> str:
        dom = xml.dom.minidom.parseString(instr)
        img = dom.documentElement
        assert img is not None
        src = img.getAttribute("src")
        alt = img.getAttribute("alt")
        return IMAGE_TEMPLATE.format(src=src, alt=alt)


class Link2Latex:
    def convert(self, instr: str) -> str:
        dom = xml.dom.minidom.parseString(instr)
        link = dom.documentElement
        assert link is not None
        href = link.getAttribute("href")
        matches = re.search(r">([^<]+)", instr)
        desc = ""
        if matches:
            desc = matches.group(1)
        return r"\href{%s}{%s}" % (href, desc) if href != desc else r"\url{%s}" % href


class ImageTextPostProcessor(markdown.postprocessors.Postprocessor):
    def run(self, instr: str) -> str:
        converter = Img2Latex()
        new_blocks = []
        for block in instr.split("\n\n"):
            stripped = block.strip()
            if stripped.startswith("<img"):
                stripped = re.sub(r"<\/?plaintext[^>]*>", "", stripped, flags=re.IGNORECASE)
                new_blocks.append(converter.convert(stripped).strip())
            else:
                new_blocks.append(block)
        return "\n\n".join(new_blocks)


class LinkTextPostProcessor(markdown.postprocessors.Postprocessor):
    def run(self, instr: str) -> str:
        converter = Link2Latex()
        new_blocks = []
        for block in instr.split("\n\n"):
            stripped = block.strip()
            matches = re.findall(r"<a[^>]*>[^<]+</a>", stripped)
            if matches:
                for match in matches:
                    stripped = stripped.replace(match, converter.convert(match).strip())
                new_blocks.append(stripped)
            else:
                new_blocks.append(block)
        return "\n\n".join(new_blocks)


class UnescapeHtmlTextPostProcessor(markdown.postprocessors.Postprocessor):
    def run(self, text: str) -> str:
        return unescape_html_entities(inline_html_latex(text))


class MathTextPostProcessor(markdown.postprocessors.Postprocessor):
    def run(self, instr: str) -> str:
        instr = re.sub(r"\$\$([^\$]*)\$\$", r"\\[\1\\]", instr)
        instr = re.sub(r"\$([^\$]*)\$", r"\\(\1\\)", instr)
        instr = instr.replace("\\lt", "<").replace(" * ", " \\cdot ").replace("\\del", "\\partial")
        return instr


class TableTextPostProcessor(markdown.postprocessors.Postprocessor):
    def run(self, instr: str) -> str:
        converter = Table2Latex()
        new_blocks = []
        for block in instr.split("\n\n"):
            stripped = block.strip()
            if re.match(r"\|.*\|", stripped):  # Check for Markdown table
                new_blocks.append(converter.convert_markdown_table(stripped).strip())
            else:
                new_blocks.append(block)
        return "\n\n".join(new_blocks)


def convert_md_to_latex(md_content: str) -> str:
    md = markdown.Markdown(extensions=[LaTeXExtension()])
    latex_content = md.convert(md_content)
    latex_content = re.sub(r"<\/?plaintext[^>]*>", "", latex_content, flags=re.IGNORECASE)
    return MAIN_TEMPLATE.format(latex_content=latex_content)


def md_to_pdf(markdown_text: str, output_filename: str = "output") -> str:
    """
    Convert Markdown to PDF via LaTeX.

    Args:
        markdown_text: Markdown text
        output_filename: Output filename (without extension)

    Returns:
        Message about the compilation result
    """

    latex_code = convert_md_to_latex(markdown_text)

    temp_dir = get_workspace_dir() / "temp_latex"
    temp_dir.mkdir(parents=True, exist_ok=True)

    tex_file_path = temp_dir / "temp.tex"
    with open(tex_file_path, "w", encoding="utf-8") as f:
        f.write(latex_code)

    try:
        subprocess.run(
            [
                "pdflatex",
                "-interaction=nonstopmode",
                "-output-directory",
                temp_dir,
                tex_file_path,
            ],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            timeout=30,
        )

    except subprocess.TimeoutExpired:
        return "Compilation timed out after 30 seconds"
    except subprocess.CalledProcessError as e:
        error_msg = e.stdout.decode("utf-8")
        error_lines = [
            line for line in error_msg.split("\n") if "error" in line.lower() or "!" in line
        ]
        if error_lines:
            return "Compilation failed. LaTeX errors:\n" + "\n".join(error_lines)
        return f"Compilation failed. Full LaTeX output:\n{error_msg}"

    pdf_path = os.path.join(temp_dir, "temp.pdf")
    output_pdf_path = os.path.join(get_workspace_dir(), f"{output_filename}.pdf")

    if os.path.exists(pdf_path):
        shutil.move(pdf_path, output_pdf_path)
        shutil.rmtree(temp_dir, ignore_errors=True)
        return f"Compilation successful! PDF file saved as {output_filename}.pdf"

    return "Compilation completed, but PDF file was not created. Check LaTeX code for errors."



import pytest

from academia_mcp.tools import arxiv_search


def test_arxiv_search_basic_search() -> None:
    result = arxiv_search('ti:"PingPong: A Benchmark for Role-Playing Language Models"')
    assert isinstance(result, str)
    assert len(result) > 0
    assert "Ilya Gusev" in result


def test_arxiv_search_empty_query() -> None:
    with pytest.raises(AssertionError):
        arxiv_search("")


def test_arxiv_search_invalid_sort_by() -> None:
    with pytest.raises(AssertionError):
        arxiv_search("abs:physics", sort_by="invalid_sort")


def test_arxiv_search_invalid_sort_order() -> None:
    with pytest.raises(AssertionError):
        arxiv_search("abs:physics", sort_order="invalid_order")


def test_arxiv_search_negative_offset() -> None:
    with pytest.raises(AssertionError):
        arxiv_search("abs:physics", offset=-1)


def test_arxiv_search_negative_limit() -> None:
    with pytest.raises(AssertionError):
        arxiv_search("abs:physics", limit=-1)


def test_arxiv_search_zero_limit() -> None:
    with pytest.raises(AssertionError):
        arxiv_search("abs:physics", limit=0)


def test_arxiv_search_offset_limit_combination() -> None:
    result1 = arxiv_search("abs:physics", offset=0, limit=5)
    result2 = arxiv_search("abs:physics", offset=5, limit=5)
    assert result1 != result2


def test_arxiv_search_sort_by_options() -> None:
    valid_sort_options = ["relevance", "lastUpdatedDate", "submittedDate"]
    for sort_option in valid_sort_options:
        result = arxiv_search("abs:physics", sort_by=sort_option)
        assert isinstance(result, str)
        assert len(result) > 0


def sort_order_options() -> None:
    result1 = arxiv_search("abs:physics", sort_order="ascending")
    result2 = arxiv_search("abs:physics", sort_order="descending")
    assert isinstance(result1, str)
    assert len(result1) > 0
    assert isinstance(result2, str)
    assert len(result2) > 0
    assert result1 != result2


def test_arxiv_search_unicode_in_query() -> None:
    result = arxiv_search('abs:"Schrödinger equation"')
    assert isinstance(result, str)
    assert len(result) > 0


def test_arxiv_search_result_format() -> None:
    result = arxiv_search("abs:physics")
    assert "title" in result.lower()


def test_arxiv_search_type_validation() -> None:
    with pytest.raises(AssertionError):
        arxiv_search(123)  # type: ignore
    with pytest.raises(AssertionError):
        arxiv_search("abs:physics", offset="0")  # type: ignore
    with pytest.raises(AssertionError):
        arxiv_search("abs:physics", limit="5")  # type: ignore
    with pytest.raises(AssertionError):
        arxiv_search("abs:physics", sort_by=123)  # type: ignore
    with pytest.raises(AssertionError):
        arxiv_search("abs:physics", sort_order=123)  # type: ignore


def test_arxiv_search_integration_multiple_pages() -> None:
    first_page = arxiv_search("abs:physics", offset=0, limit=5)
    second_page = arxiv_search("abs:physics", offset=5, limit=5)
    third_page = arxiv_search("abs:physics", offset=10, limit=5)
    assert first_page != second_page != third_page
    assert isinstance(first_page, str)
    assert isinstance(second_page, str)
    assert isinstance(third_page, str)


def test_arxiv_search_date_filter() -> None:
    result = arxiv_search("au:vaswani", start_date="2017-06-01", end_date="2017-07-01", limit=2)
    assert "Attention Is All You Need" in result


def test_arxiv_search_start_date_only() -> None:
    result = arxiv_search('au:vaswani AND ti:"attention is all"', start_date="2017-06-01")
    assert "Attention Is All You Need" in result


def test_arxiv_search_end_date_only() -> None:
    result = arxiv_search('au:vaswani AND ti:"attention is all"', end_date="2017-07-01")
    assert "Attention Is All You Need" in result


def test_arxiv_search_complex_query_vq_vae() -> None:
    result = arxiv_search(
        query='abs:"VQ-VAE" OR abs:"Vector Quantized Variational Autoencoders"',
        limit=1,
        sort_by="submittedDate",
        sort_order="ascending",
        include_abstracts=True,
    )
    assert "1711.00937" in result


def test_arxiv_search_point_e() -> None:
    result = arxiv_search(
        query='ti:"Point-E"',
        limit=1,
        sort_by="submittedDate",
        sort_order="ascending",
        include_abstracts=True,
    )
    assert "2212.08751" in result



# https://arxiv.org/abs/2504.12976
# https://web.stanford.edu/class/cs197c/slides/02-literature-search.pdf

import json
import os
import random
from typing import List, Optional, Any

from pydantic import BaseModel
from datasets import load_dataset  # type: ignore

from academia_mcp.tools.arxiv_download import arxiv_download
from academia_mcp.utils import extract_json, encode_prompt
from academia_mcp.llm import llm_acall


class ProposalDataset:
    dataset: Optional[List[Any]] = None

    @classmethod
    def get_dataset(cls) -> List[Any]:
        if cls.dataset is None:
            cls.dataset = list(load_dataset("UniverseTBD/hypogen-dr1")["train"])
        return cls.dataset


EXTRACT_PROMPT = """
You are a highly advanced research assistant.
You specialize in reading scientific papers for hypothesis generation and identifying innovative ideas.


## Example (BERT in NLP)
Before you begin, let 's revisit the Bit-Flip concept with an example (BERT in NLP):
- Bit: Traditional NLP models (RNNs, LSTMs) process text sequentially,
limiting their ability to understand long-range dependencies and fully capture bidirectional context.
- Flip: Instead, consider entire sentences at once, allowing context from both directions. This helps capture nuanced relationships among words.
- Spark: Bidirectional context for NLP.

## Framework
A Bit-Flip inverts a commonly held assumption,
questioning existing constraints or reapplying techniques to new domains/scales.
The "Bit" is the prevailing belief, and the "Flip" is the counterargument.

## Guidance for analysis
1. Bit (Technical Insight):
- Provide at least two sentences clearly stating the status quo or conventional approach.
- Highlight the limitation or problem it creates.
- Include enough detail so it is self-contained and does not rely on additional context from elsewhere.
2. Flip (Innovation):
- Provide at least two sentences describing the novel approach or perspective.
- Explain the method or technique that enables this change.
- Include enough detail so the Flip is understandable on its own.
3. Spark (Core Summary):
- A concise 4-6 word phrase capturing the core idea.

Now, consider this research abstract:
{{abstract}}

Your task:
Identify the Bit, Flip, and Spark from the abstract in a detailed manner:
- Bit: at least two sentences, with sufficient detail about the conventional approach and its limitation.
- Flip: at least two sentences, describing the new approach or perspective with enough detail to understand the main technique.
- Spark: a concise 4-6 word summary of the core idea.

Follow these rules:
- Do not cite the paper itself or its authors.
- Instead of saying "We/I introduced an idea", just say "An idea was introduced ...".

Return only the JSON object in this exact format (no extra text):
{
    "bit": "Technical limitation or conventional approach, in at least two sentences",
    "flip": "Innovative approach or solution, in at least two sentences",
    "spark": "4-6 word summary"
}
"""

IMPROVEMENT_PROMPT = """
You are a highly advanced research assistant.
You specialize in hypothesis generation and identifying innovative ideas.

You are given a Bit, which is a technical limitation or conventional approach of some paper.
Your task is to propose an improvement idea for the Bit called Flip and summarize it in a Spark.
Do not propose any human annotations or human-in-the-loop, the idea should be automatically verifiable.
Try to be as specific as possible.

{% for example in examples %}
## Example {{loop.index}}
- Bit: {{example["bit"]}}
- Chain of reasoning: {{example["chain_of_reasoning"]}}
- Flip: {{example["flip"]}}
- Spark: {{example["spark"]}}
{% endfor %}

Now, please propose a chain of reasoning that leads to an improvement idea for this Bit:
{{bit}}

{% if additional_context %}Additional context:
{{additional_context}}{% endif %}

Finalize your idea by providing the idea details:
- Abstract: An abstract that summarizes the proposal in conference format (approximately 250 words).
- Experiments: A list of experiments that would be conducted to validate the proposal. Ensure these are simple and feasible. Be specific in exactly how you would test the hypothesis, and detail precise algorithmic changes. Include the evaluation metrics you would use.
- Risks and limitations: A list of potential risks and limitations of the proposal.

Return only the JSON object in this exact format (no extra text):
{
    "chain_of_reasoning": "Chain of reasoning that leads to an improvement idea for this Bit. At least 5 sentences.",
    "flip": "Innovative approach or solution, in at least two sentences",
    "spark": "4-6 word summary",
    "abstract": "An abstract that summarizes the proposal in conference format (approximately 250 words).",
    "experiments": ["...", "..."],
    "risks_and_limitations": "A list of potential risks and limitations of the proposal."
}
"""


SCORE_PROMPT = """
You are a highly advanced research assistant.
You are given a list of research proposals.
Your task is to score the proposals.

Proposals:
{% for proposal in proposals %}
- Proposal ID: {{proposal["proposal_id"]}}
- Spark: {{proposal["spark"]}}
- Abstract: {{proposal["abstract"]}}
- Experiments: {{proposal["experiments"]}}
- Risks and limitations: {{proposal["risks_and_limitations"]}}
{% endfor %}

Here are the criteria:
- "Strengths": A list of strengths of the proposal.
- "Weaknesses": A list of weaknesses of the proposal.
- "Novelty": Is the proposal novel? A rating from 1 to 4 (low, medium, high, very high).
- "Clarity": Is the proposal clear? A rating from 1 to 4 (low, medium, high, very high).
- "Significance": Is the proposal significant? A rating from 1 to 4 (low, medium, high, very high).
- "Feasibility": Is the proposal feasible and easy to implement? A rating from 1 to 4 (low, medium, high, very high).
- "Soundness": Is the proposal sound? A rating from 1 to 4 (poor, fair, good, excellent).
- "Overall": A rating from 1 to 10 (very strong reject to award quality).

Return only scores for all proposals in this exact format (no extra text):
[
    {
        "proposal_id": 0,
        "spark": "...",
        "strengths": ["...", "..."],
        "weaknesses": ["...", "..."],
        "novelty": 2,
        "clarity": 2,
        "significance": 2,
        "feasibility": 2,
        "soundness": 2,
        "overall": 5
    },
    ...
]
"""


class BitFlipInfo(BaseModel):  # type: ignore
    bit: str
    flip: str
    spark: str


class Proposal(BaseModel):  # type: ignore
    proposal_id: Optional[int] = None
    flip: str
    spark: str
    abstract: str
    experiments: List[str]
    risks_and_limitations: List[str]


class ProposalScores(BaseModel):  # type: ignore
    proposal_id: int
    spark: str
    strengths: List[str]
    weaknesses: List[str]
    novelty: int
    clarity: int
    significance: int
    feasibility: int
    soundness: int
    overall: int


async def extract_bitflip_info(arxiv_id: str) -> str:
    """
    Extracts the Bit-Flip information from the arXiv paper.

    A Bit-Flip is a technique that inverts a commonly held assumption,
    questioning existing constraints or reapplying techniques to new domains/scales.
    The "Bit" is the prevailing belief, and the "Flip" is the counterargument.

    Returns a JSON object in this format:
    {
        "bit": "Technical limitation or conventional approach, in at least two sentences",
        "flip": "Innovative approach or solution, in at least two sentences",
        "spark": "4-6 word summary of the core idea"
    }
    Use `json.loads` to deserialize the result if you want to get specific fields.

    Args:
        arxiv_id: The arXiv ID of the paper to extract the Bit-Flip information from.
    """
    model_name = os.getenv("BITFLIP_MODEL_NAME", "deepseek/deepseek-chat-v3-0324")
    paper = arxiv_download(arxiv_id)
    abstract = json.loads(paper)["abstract"]
    prompt = encode_prompt(EXTRACT_PROMPT, abstract=abstract)
    content = await llm_acall(model_name=model_name, prompt=prompt)
    result = extract_json(content)
    bitflip_info: BitFlipInfo = BitFlipInfo.model_validate(result)
    return str(bitflip_info.model_dump_json())


async def generate_research_proposal(bit: str, additional_context: str = "") -> str:
    """
    Proposes an improvement idea for the Bit.

    Args:
        bit: The Bit to propose an improvement idea for. The bit is a technical limitation or conventional approach of some paper.
        additional_context: Additional context to use when proposing the improvement idea.

    Returns a JSON string with a research proposal in this format:
    {
        "proposal_id": ...,
        "flip": "Innovative approach or solution, in at least two sentences",
        "spark": "4-6 word summary",
        "abstract": "An abstract that summarizes the proposal in conference format (approximately 250 words).",
        "experiments": ["...", "..."],
        "risks_and_limitations": "A list of potential risks and limitations of the proposal."
    }
    Use `json.loads` to deserialize the result if you want to get specific fields.
    """
    model_name = os.getenv("BITFLIP_MODEL_NAME", "deepseek/deepseek-chat-v3-0324")
    examples = ProposalDataset.get_dataset()[:]
    examples = random.choices(examples, k=4)

    prompt = encode_prompt(
        IMPROVEMENT_PROMPT, bit=bit, examples=examples, additional_context=additional_context
    )
    content = await llm_acall(model_name=model_name, prompt=prompt)
    result = extract_json(content)
    proposal: Proposal = Proposal.model_validate(result)
    proposal.proposal_id = random.randint(0, 1000000)
    return str(proposal.model_dump_json())


async def score_research_proposals(proposals: List[str]) -> str:
    """
    Scores a list of research proposals.
    Use proposals obtained with the `generate_research_proposal` tool.

    Returns a JSON string with a list of scores in this format:
    [
        {
            "proposal_id": 0,
            "spark": "...",
            "strengths": ["...", "..."],
            "weaknesses": ["...", "..."],
            "novelty": 2,
            "clarity": 2,
            "significance": 2,
            "feasibility": 2,
            "soundness": 2,
            "overall": 5
        },
        ...
    ]
    Use `json.loads` to deserialize the result if you want to get specific fields.

    Args:
        proposals: A list of JSON strings with research proposals.
    """
    model_name = os.getenv("BITFLIP_MODEL_NAME", "deepseek/deepseek-chat-v3-0324")
    proposals = [Proposal.model_validate_json(proposal) for proposal in proposals]
    prompt = encode_prompt(SCORE_PROMPT, proposals=proposals)
    content = await llm_acall(model_name=model_name, prompt=prompt)
    scores = extract_json(content)
    final_scores = [ProposalScores.model_validate(score) for score in scores]
    return json.dumps([s.model_dump() for s in final_scores], ensure_ascii=False)



import re
import json
from urllib3.util.retry import Retry
from typing import Dict, Any, Optional

import requests
from jinja2 import Template


def post_with_retries(
    url: str,
    payload: Dict[str, Any],
    api_key: Optional[str] = None,
    timeout: int = 30,
    num_retries: int = 3,
) -> requests.Response:
    retry_strategy = Retry(
        total=num_retries,
        backoff_factor=3,
        status_forcelist=[429, 500, 502, 503, 504],
        allowed_methods=["POST"],
    )

    session = requests.Session()
    adapter = requests.adapters.HTTPAdapter(max_retries=retry_strategy)
    session.mount("https://", adapter)

    headers = {
        "x-api-key": api_key,
        "x-subscription-token": api_key,
        "Authorization": f"Bearer {api_key}",
        "Content-Type": "application/json",
    }

    response = session.post(url, headers=headers, json=payload, timeout=timeout)
    response.raise_for_status()
    return response


def get_with_retries(
    url: str,
    api_key: Optional[str] = None,
    timeout: int = 30,
    num_retries: int = 3,
    params: Optional[Dict[str, Any]] = None,
) -> requests.Response:
    retry_strategy = Retry(
        total=num_retries,
        backoff_factor=30,
        status_forcelist=[429, 500, 502, 503, 504],
        allowed_methods=["GET"],
    )

    session = requests.Session()
    adapter = requests.adapters.HTTPAdapter(max_retries=retry_strategy)
    session.mount("https://", adapter)

    headers = {}
    if api_key:
        headers["x-api-key"] = api_key
        headers["x-subscription-token"] = api_key
        headers["Authorization"] = f"Bearer {api_key}"

    response = session.get(url, headers=headers, timeout=timeout, params=params)
    response.raise_for_status()
    return response


def clean_json_string(text: str) -> str:
    try:
        return json.dumps(json.loads(text))
    except json.JSONDecodeError:
        pass
    text = text.strip()
    text = re.sub(r",(\s*[}\]])", r"\1", text)
    text = re.sub(r"'([^']*)':", r'"\1":', text)
    text = re.sub(r":\s*'([^']*)'", r': "\1"', text)
    text = re.sub(r"//.*?$", "", text, flags=re.MULTILINE)
    text = re.sub(r"/\*.*?\*/", "", text, flags=re.DOTALL)

    prefixes_to_remove = [
        "json:",
        "JSON:",
        "Here is the JSON:",
        "Here's the JSON:",
        "The JSON is:",
        "Result:",
        "Output:",
        "Response:",
    ]

    for prefix in prefixes_to_remove:
        if text.lower().startswith(prefix.lower()):
            text = text[len(prefix) :].strip()

    return text


def extract_json(text: str) -> Any:
    assert isinstance(text, str), "Input must be a string"

    text = text.strip()
    assert text, "Input must be a non-empty string"

    json_blocks = re.findall(r"```json\s*(.*?)\s*```", text, re.DOTALL | re.IGNORECASE)
    for block in json_blocks:
        try:
            return json.loads(block.strip())
        except json.JSONDecodeError:
            continue

    code_blocks = re.findall(r"```\s*(.*?)\s*```", text, re.DOTALL)
    for block in code_blocks:
        block = block.strip()
        if block.startswith(("{", "[")):
            try:
                return json.loads(block)
            except json.JSONDecodeError:
                continue

    try:
        return json.loads(clean_json_string(text))
    except json.JSONDecodeError:
        pass

    json_patterns = [
        r"\{[^{}]*(?:\{[^{}]*\}[^{}]*)*\}",
        r"\[[^\[\]]*(?:\[[^\[\]]*\][^\[\]]*)*\]",
        r"\{.*\}",
        r"\[.*\]",
    ]

    for pattern in json_patterns:
        matches = re.findall(pattern, text, re.DOTALL)
        for match in sorted(matches, key=len, reverse=True):
            try:
                cleaned = clean_json_string(match.strip())
                return json.loads(cleaned)
            except json.JSONDecodeError:
                continue

    return None


def encode_prompt(template: str, **kwargs: Any) -> str:
    template_obj = Template(template)
    return template_obj.render(**kwargs).strip()
